/* Aine knowledge files compiler

 License: GPL
 This is part of the Aine Project.
 For copyrights and license (GPL) - see main.c in main source dir
 
*/

#include <stdarg.h>
#include "aine_c.h"

#undef FALSE
#undef TRUE
typedef enum { FALSE, TRUE } bool;

#ifdef ftell
#define TPOS ftell(tpos)
#else
#ifdef __linux__
#define TPOS tpos.__pos
#else
#define TPOS tpos
#endif
#endif

FILE *indexpos_f, *inputs_a_f, *inputs_a_t, *inputs_d_f, *inputs_d_t, *aine;

fpos_t chunk[53]; /* 26 * 2 + 1 */

char *topic;   /* Topic holds our topic which we convert to conditional tags */
short topic_count;
FILE *index_f, *topicfile, *unsorted_a_f, *unsorted_a_t, *unsorted_d_f,
*unsorted_d_t, *outputs_f;
bool comment = FALSE;
bool output_check;
unsigned long size, multi_inputs; /* number of categories bot have */


short str_len_aine (char *s)
{
	int count=0;
	while (*s != '}' && *s != '\0')
		s++;
	if (*s == '\0')
		return 0;
	while (*s != '<' && *s != '\0') {
		count++; s++; 
	}
	return count;
}


short str_len_up_to_that (char *s)
{		/* and without topic */
	if (!strchr (s, '<')) return 0; 
	return ( strchr (s, '<') - s);
}

void binary_writer (FILE *out, char *s)
{		/* quick and nasty fix for binary output */
	char *p;
	int i, j;
	char *buf = (char*) malloc (strlen(s) * sizeof(char));
	if (!buf) exit (12);
	
	if ((p=strchr (s, '\n'))) *p = '\0';
	p = s;
	while (*(++p) != '}') ;
	*p = '\0';
	i =  atoi (s);
	fwrite (&i, sizeof (int), 1, out);
	i = (int) strlen (++p);
	strcpy (buf, p);
	*strchr (buf, '&') = '\0';
	i++;
	fwrite (&i, sizeof (int), 1, out);
	j = (int) strlen (buf);
	fwrite (&j, sizeof (int), 1, out);
	fwrite (p, sizeof(char), (size_t) i, out);
	free (buf);
}


void replacer (char *line)
{
	char *pos;
	short i;
	
	if (line[0] == '\0' || line[1] == '\0')
		return;
  	if (comment && strstr (line, "-->")) { 
		line[0] = '\0';
		comment = FALSE;
	}
	if (strstr (line, "<!--") && strstr (line, "-->"))
		line[0] = '\0';
  	if (strstr (line, "<!--")) comment = TRUE;
	if (comment)
		line[0] = '\0';
	replace (line, "<R>", "<r><l>");
	replace (line, "</R>", "</l></r>");
	replace_all (line, "<*=0>", "<*>");
	replace_all (line, "<star/>", "<*>");
	replace_all (line, "{*}", "{<*>}");
	replace_all (line, "<br/>", "<br>");

	/* next 3 are rather only for my (Polish) bot...
	   I don't think you will need them */
	replace_all (line, "<plec>", "<c=gender><k=on>");
	replace_all (line, "<ona>", "</k><k>");
	replace_all (line, "</plec>", "</k></c>"); 

	if ((pos = strstr (line, "<topic="))) {
		topic_count++;
		if (topic_count > 999) {
			fputs ("ERROR - more than 1000 topics are not allowed", stderr);
			exit (1);
		}
		if (pos[7] == '\"') pos++; 
		strcpy (topic, pos + 7);
		*strchr (topic,'>') = '\0';
		replace (topic, "\"", "");
		uppercase (topic);
		i = strlen (topic) + 1;
		fwrite (&i, sizeof(short), 1, topicfile);
		fwrite (topic, sizeof(char), i, topicfile);
		fwrite (&topic_count, sizeof(short), 1, topicfile);
	}
	if (strstr (line, "</topic>"))
		topic[0] = '\0';
	strremove (line, "<topic=", ">");
	replace (line, "</topic>", "");
	replace_all (line, "\t","");
	replace_all (line, "\r","");
	replace_all (line, "\n","");
	replace_all (line, "<input/>", "<input=1>");
	replace_all (line, "<input index=", "<input=");
	replace_all (line, "<person/>", "<person><*></person>");
	replace_all (line, "<person2/>", "<person2><*></person2>");
	replace_all (line, "<personf/>", "<personf><*></personf>");
	replace_all (line, "<sr/>", "{<*>}");
	replace_all (line, "</*>", "{<*>}");
}


int splitaiml (char *src)
{
	unsigned long size_this = 0;
	int patterns = 0, templates = 0;
	fpos_t tpos;
	char *line, *next;
	register int linenr = 0;
	int atomic;
	char *question;
	char response[(sizeof(char)*MAX_LINE_SIZE)];
	
	char *tokenPtr;
	int that_used = 0;
	char *that_string = NULL;
	char that_line[MAX_LINE_SIZE];
	unsigned long multi_inputs_this = 0;
	unsigned short int i, j;
	
	aine = fopen (src, "r+");
	if (!aine) {
		printf ("%s not found, skipping!\n", src);
		return (1);
	} else {
		printf ("Processing: %s\n", src);
	}
	
	/* here check last two lines if they are blank lines */
	fseek (aine, -3, SEEK_END);
	for (j=0, i=0; i < 3; i++) {
		if (getc(aine) != '\n')
			j++;
	}
	/* we put the lines if the user forget */
	if (j != 0) {
		for (i = 0; i < j; i++)
			fputc ('\n', aine);
	}
	/* check syntax */
  	printf ("Checking %s ...", src);
	fflush (stdout);
	check_aine_language (aine);
  	
	/* fill dictionary */
	printf ("Added %d words to dictionary\n", dictionary (aine));
		
	line = (char*) malloc (sizeof(char)*MAX_LINE_SIZE);
	next = (char*) malloc (sizeof(char)*MAX_LINE_SIZE);
	question = (char*) malloc (sizeof(char)*MAX_LINE_SIZE);
	
	assert (line);
	assert (next);
	assert (question);
	memset (line, 0, (sizeof(char)*MAX_LINE_SIZE));
	memset (next, 0, (sizeof(char)*MAX_LINE_SIZE));
	memset (question, 0, (sizeof(char)*MAX_LINE_SIZE));
	
	while (fgets (line, MAX_LINE_SIZE, aine)) {
		/* run replacer first to deal with 'topic' lines  */
		replacer (line);
		
		if (strstr (line, "+[")) {
			linenr++;
			size++;
			size_this++;
			patterns++;
			
			if (strstr (line, "<t>")) {
				while (!strstr (line, "</t>")) {
					fgets (next, MAX_LINE_SIZE, aine);
					replacer (next);
					strcpy (line + strlen (line), next);
					linenr++;
				}
				that_used = 1;
			} else {
				/* second check for '<t>' could go here */
				while (!strchr (line, ']')) {
					fgets (next, MAX_LINE_SIZE, aine);
					replacer (next);
					strcpy (line + strlen (line), next);
					linenr++;
					*(strchr (line, ']') + 1) = '\0';
				}
				that_used = 0;
			}
			
			/* 'escape' punctuation marks with a '&' */
			replace (line, ".", "&.");
			replace (line, "?", "&?");
			replace (line, "!", "&!");
			
			/* copy line as question and truncate the 'that' part*/
			strcpy (question, line);
			if (strstr (question, "<t>"))
				*strstr (question,"<t>") = '\0';
			replace_all (question, "\n", "");
			replace (question, "]", "");
			replace (question, "+[", "");
			
			/* extract the 'that' part */
			if (that_used == 1) {
				that_string = strstr (line, "<t>");
				replace (that_string, "<t>", "");
				replace (that_string, "</t>", "");
				strcpy (that_line, that_string);
				replace (that_line, "\n", "\0");
			}
			
			/* Now get the  response */
			fgets (line, MAX_LINE_SIZE, aine);
			linenr++;
			replacer (line);
			
			while (!strstr (line, "-[")) {
				fgets (next, MAX_LINE_SIZE, aine);
				strcpy (line + strlen (line), next);
				replacer (line);
				linenr++;
			}
			
			while (!strchr (line, ']')) {
				fgets (next, MAX_LINE_SIZE, aine);
				strcpy (line + strlen (line), next);
				replacer (line);
				linenr++;
			}
			
			templates++;
			
			/* strip unneeded brackets */
			replace_all (line, "\n", "");
			*strchr (line, ']') = '\0';
			replace (line, "-[", "");
			
			fgetpos (outputs_f, &tpos);
			fprintf (outputs_f, "%s\n", line);
			
			/* copy line as  'response' so tokenPtr is more readable */
			strcpy (response, line);
			
			/* tokenize using the pipe '|' as separator*/
			/* +[TEST6|TESTME] <t>\OK5</t>*/
			/* -[multi-input atomic with that] */
			tokenPtr = strtok( question, "|" ); 
			
			while ( tokenPtr != NULL ) { 
				multi_inputs_this++;
				replace (tokenPtr, "\n", "");
				
				/* check each chunk for atomicity */
				/* +[TESTNOTEQUAL|TESTINEQUAL2|FUNNYSTUFF *] */
				/* -[multi-input atomic and default inputs] */
				if (strchr (tokenPtr, '*') || strchr (tokenPtr, '_')) {
					atomic = 0;
				} else {
					atomic = 1;
				}
				
				if (that_used==1) {
					if (topic[0] == 0) {
						if ((atomic==1)) { 
							fprintf (unsorted_a_t, "%s&<%s", tokenPtr, that_line);
						} else { 
							fprintf (unsorted_d_t, "%s&<%s", tokenPtr, that_line);
						}
					} else {
						if ((atomic==1)) {
							fprintf (unsorted_a_t, "%s&<%s", tokenPtr, that_line);
						} else {
							fprintf (unsorted_d_t, "%s&<%s", tokenPtr, that_line);
						}
					}
				} else {
					if (topic[0] == 0) {
						if ((atomic==1)) { 
							fprintf (unsorted_a_f, "%s&", tokenPtr);
						} else { 
							fprintf (unsorted_d_f, "0}%s&", tokenPtr);
						}
					} else {
						if ((atomic==1)) { 
							fprintf (unsorted_a_f, "%s&", tokenPtr);
						} else {
							fprintf (unsorted_d_f, "%d}%s&", topic_count, tokenPtr);
						}
					}
				}
				
				/* finish writing the unsorted files */
				if (atomic==1) {
#if defined (__linux__) || (__sun)
					if (that_used==1) fprintf (unsorted_a_t, "<$=%ld\n",  TPOS); 
					else fprintf (unsorted_a_f, "<$=%ld\n", TPOS);
#else
					if (that_used==1) fprintf (unsorted_a_t, "<$=%lld\n",  TPOS); 
					else fprintf (unsorted_a_f, "<$=%lld\n", TPOS);
#endif
				} else {
#if defined (__linux__) || (__sun)
					if (that_used==1) 	fprintf (unsorted_d_t, "<$=%ld\n", TPOS);  
					else fprintf (unsorted_d_f, "<$=%ld\n", TPOS);
#else
					if (that_used==1) 	fprintf (unsorted_d_t, "<$=%lld\n", TPOS);
					else fprintf (unsorted_d_f, "<$=%lld\n", TPOS);
#endif
				}
				
				/* advance to the next token */
				tokenPtr = strtok( NULL, "|" );
			}
		}
		/* strcpy (line, next); */
		that_used = 0;
		that_string = NULL;
	}
	
	printf ("Added %lu patterns and %lu multi-input patterns\n\n", size_this, multi_inputs_this - size_this);
	multi_inputs += (multi_inputs_this - size_this);
	multi_inputs_this = 0;
	
	/* we're outta here */
	fclose (aine);
	free(line);
	free(next);
	free(question);
	
  return 1;
}



int prepare (void)
{
	register short i;
	short ind;
	char *aline = (char*) malloc (sizeof (char) * 256);
	char *buffer = (char*) malloc (sizeof (char) * MAX_LINE_SIZE);
	
	topic = (char*) malloc (sizeof(char)*MAX_VARVAL_SIZE);
  
	output_check = FALSE;
  
	puts ("Loading index file of AINE brain files:\n");
  
	while (fgets (aline, 256, index_f)) {
		aline[strcspn (aline, "\n")] = 0;
		if (!splitaiml (aline))
			return (0);
		memset (aline, 0, 256);
	}
	free (aline);

  	/*   S O R T I N G    */
  
		/* SORTING "ATOMIC *without* THAT" */
	puts ("Sorting patterns. Please be patient.");
	printf ("Sorting unsorted_a_f: ");
	fclose (unsorted_a_f);
	unsorted_a_f = fopen ("data/brain/unsorted_a_f", "r");
	for (i = 'A'; i <= 'Z'; i++) {
		putchar (i);
		fgetpos (inputs_a_f, &chunk[i- 'A']);
		for (ind = 0; ind < 2; ind++) { /* if (ind == 0) we only printf patterns with
a ? or ! or . So these are always first. */
			rewind (unsorted_a_f);
			while (fgets (buffer, MAX_LINE_SIZE, unsorted_a_f)) {
				if (!ind && buffer[0] == i && (strchr (buffer, '.') || strchr (buffer, '!')
|| strchr (buffer, '?'))) {
					fputs (buffer, inputs_a_f);
					continue;
				}
				if (ind && buffer[0] == i && (!strchr (buffer, '.') && !strchr (buffer,
'!') && !strchr (buffer, '?'))) {
					fputs (buffer, inputs_a_f);
					continue;
				}
				if (ind && i == 'Z' && (buffer[0] < 'A'  ||  buffer[0] > 'Z'))
					fprintf (unsorted_d_f, "0}%s", buffer);
			}			
		}
	}
	
	fputs ("eof&<$=0\n", inputs_a_f);
	/* we need second 'eof' as well (loop unroll in matchers.c) */
	fputs ("eof2&<$=0", inputs_a_f);
	
	putchar ('\n');  
	
	/* SORTING "ATOMIC WITH THAT" */
	fclose (unsorted_a_t);
	unsorted_a_t = fopen ("data/brain/unsorted_a_t", "r");
	printf ("Sorting unsorted_a_t: ");
	for (i = 'A'; i <= 'Z'; i++) {
		putchar (i);
		fgetpos (inputs_a_t, &chunk[i - 'A' + 26]);
		for (ind=0; ind < 2; ind++) {
			rewind (unsorted_a_t);
			while (fgets (buffer, MAX_LINE_SIZE, unsorted_a_t)) {
				if (!ind && buffer[0] == i && (strchr (buffer, '.')
				  || strchr (buffer, '!') || strchr (buffer, '?'))) {
					fputs (buffer, inputs_a_t);
					continue;
				}
				if (ind && buffer[0] == i && (!strchr (buffer, '.')
				  && !strchr (buffer, '!') && !strchr (buffer, '?'))) {
					fputs (buffer, inputs_a_t);
					continue;
				}
				if (ind && i == 'Z' && (buffer[0] < 'A'  || buffer[0] > 'Z'))
					fputs (buffer, unsorted_d_t);
			}
	
		}
	}

#ifdef __linux__
#define CHUNK &chunk[ind].__pos
#else
#define CHUNK &chunk[ind]
#endif
	
	fputs ("eof&<Nothing<$=0\n", inputs_a_t);
	fputs ("eof2&<Nothing<$=0", inputs_a_t); /* we need two eof's (loop unroll) */
	putchar ('\n');
	ind = 0;
	while (ind < 53) { /* 26*2 letters of alphabet */
		fwrite (CHUNK, sizeof(fpos_t), 1, indexpos_f);
		ind++;
	}
	fclose (unsorted_d_f);
	printf ("Sorting unsorted_d_f. Please wait: ");
  	
	/* SORTING "DEFAULT *without* THAT" (but with topic!) */
	unsorted_d_f = fopen ("data/brain/unsorted_d_f", "r");
	for (i = 80; i > 0; i--) {
		/* first put topic0: */
		rewind (unsorted_d_f);
		while (fgets (buffer, MAX_LINE_SIZE, unsorted_d_f)) {
			if (buffer[0] == '0') {
				if (i == 80  && (str_len_aine(buffer) >= 80 || strchr(buffer, '_')
				 || (!strchr(buffer, '*') && !strchr(buffer, '_')))) { 
					replace_all(buffer, "_", "*");
					binary_writer(inputs_d_f, buffer);
				}
				else {
					if (str_len_aine(buffer) == i && strchr(buffer, '*') && !strchr(buffer, '_')) {
						binary_writer(inputs_d_f, buffer);
					}
				}
			}
			
		}
		/* then rest of topics: */
		rewind (unsorted_d_f);
		while (fgets (buffer, MAX_LINE_SIZE, unsorted_d_f)) {
			if (buffer[0] != '0') {
				if (i == 80 && (str_len_aine(buffer) >= 80 || strchr(buffer, '_')) ) {
					replace_all (buffer, "_", "*");
					binary_writer (inputs_d_f, buffer);
				}
				else {
					if (str_len_aine(buffer) == i && !strchr(buffer, '_')) {
						binary_writer (inputs_d_f, buffer);
					}
				}
			}
		}
	}
  	puts ("OK");
	fclose (unsorted_d_t);

    	/* SORTING "DEFAULT WITH THAT" */
	printf ("Sorting unsorted_d_t: ");
	unsorted_d_t = fopen ("data/brain/unsorted_d_t", "r");
	for (i = 80; i > 0; i--) {
		rewind (unsorted_d_t);
		while (fgets (buffer, MAX_LINE_SIZE, unsorted_d_t)) {
			if (i == 80 && (str_len_up_to_that (buffer) >= 80 || !strchr (buffer, '*')))   
		      		fputs (buffer, inputs_d_t);
			else {
				if (str_len_up_to_that (buffer) == i && strchr (buffer, '*'))   
					fputs (buffer, inputs_d_t);
      			}
    		}
  	}
	free (buffer);
	free (topic);
	return (1);
}

